/************************************************************************/
/* Image2Code.c - Converts image-files into static data structures 	*/
/*	in C-code, that can be used to render images directly from 	*/
/*	self-contained C-programs, without referencing external 	*/
/*	image-files at run-time.  Simple but reasonably efficient.	*/
/*									*/
/* Usage:								*/
/*	image2code image_file(s)					*/
/*									*/
/* Usage Examples:							*/
/*	image2code your_image.pnm					*/
/*	   (Creates tmp_data.c by default.  Include it in C programs, 	*/
/*	    to render the image without the original file.)		*/
/*									*/
/*   At the bottom of the produced file is a comment that shows how to	*/
/*   render the image in the captured data structure.			*/
/*									*/
/* Options:								*/
/*	-o xx	 - Send output to file xx instead of tmp_data.c.	*/
/*	-qual xx - Sets quality level to xx.  Allowable values range 	*/
/*		   from 3 down to 0.  Default quality level is 3, which */
/*		   preserves good quality for all images with reasonable*/
/*		   code size.  A value of 2 reduces code size, with	*/
/*		   usually little loss in image quality.  A value	*/
/*		   of 1 reduces code size further, and may be OK for	*/
/*		   line drawings.  Quality of zero produces the most	*/
/*		   compact code tables.					*/
/*									*/
/* Compile on Linux or Mac-OSx with:					*/
/*	cc -O -I/usr/X11R6/include -L/usr/X11R6/lib image2code.c \	*/
/*		-lGLU -lGL -lXmu -lXext -lX11 -lm -o image2code		*/
/*									*/
/* Compile on Microsoft with MinGW:					*/
/*	gcc -O -Lc:\MinGW\lib -Ic:\MinGW\include image2code.c -lglu32 \	*/
/*		-lopengl32 -lwinmm -lgdi32 -lws2_32 -lm -o image2code	*/
/*									*/
/*									*/
/************************************************************************/


#include "otk_lib/otk_lib.c"


unsigned char byte_encode( unsigned char byte )
{ /* Reduce raw 8-bit bytes (0-255) to one of 64 (6-bit) printable safe characters. */
 byte = byte >> 2;			/* First reduce to one of 64 values (6-bits). */
 if (byte < 10) return byte + 48;	/* For values 0-9, use the ascii characters "0-9". */
 if (byte < 36) return byte + 55;	/* For values 10-35, use the ascii characters "A-Z". */
 else return byte + 61;			/* For values 36-63, use the ascii characters "a-z,{,|". */
}


unsigned char byte_decode( unsigned char byte )
{ /* Reverse what byte_encode did, within available precision. */
 if (byte < 65) return ((byte - 48) << 2) + 1;
 if (byte < 97) return ((byte - 55) << 2) + 1;
 else return ((byte - 61) << 2) + 1;
}

unsigned char byte_encode64( unsigned char byte )
{ /* Reduce raw 6-bit quantities (0-63) to one of 64 printable safe characters. */
 if (byte < 10) return byte + 48;	/* For values 0-9, use the ascii characters "0-9". */
 if (byte < 36) return byte + 55;	/* For values 10-35, use the ascii characters "A-Z". */
 else return byte + 61;			/* For values 36-63, use the ascii characters "a-z,{,|". */
}


unsigned char byte_decode64( unsigned char byte )
{ /* Reverse what byte_encode64 did. */
 if (byte < 65) return (byte - 48);
 if (byte < 97) return (byte - 55);
 else return (byte - 61);
}




struct Otk_image *I2C_Read_Image_File( char *fname )
{
 struct Otk_image *image;
 FILE *infile;
 char buf[2048], capsfname[2048], *ofname;
 int ispipe=0, istmp=0;
 int magic[2], ch;
 int nrows, ncols, maxval, row, col=0, r, g, b;
 unsigned char rc, bc, gc;

 ofname = fname;
 do {capsfname[col] = toupper(fname[col]); col++; } while (fname[col-1]!='\0');

  /* Check if image needs to be converted. */
  if ((strstr(capsfname,".JPG")!=0) || (strstr(capsfname,".JPEG")!=0))
   {
    sprintf(buf,"jpegtopnm %s > tmpimg.ppm", fname);
    system(buf);  fname = strdup("tmpimg.ppm");   istmp = 1;
   }
  else
  if (strstr(capsfname,".GIF")!=0) 
   {
    sprintf(buf,"giftopnm %s > tmpimg.ppm", fname);
    system(buf);  fname = strdup("tmpimg.ppm");   istmp = 1;
  }
 else
 if (strstr(capsfname,".PNG")!=0) 
  {
   sprintf(buf,"pngtopnm %s > tmpimg.ppm", fname);
   system(buf);  fname = strdup("tmpimg.ppm");    istmp = 1;
  }

 infile = fopen(fname,"rb");
 if (infile==0) { printf("ERROR: Could not open '%s'.\n", fname);  return 0; }
 if (Otk_verbose) printf("READING IMAGE %s\n",fname);
 magic[0] = getc(infile);  /* Read magic numerals. */
 magic[1] = getc(infile);
 if ( magic[0] != 'P' || magic[1] < '1' || magic[1] > '6' ) 
  {
   fclose( infile );
   snprintf( buf, 2048, "anytopnm %s", fname );
   infile = popen( buf, "rb" );
   if (infile==0) { printf("OTK ERROR: Could not read or convert '%s'.\n", fname);  return 0; }
   ispipe = 1;
   magic[0] = getc(infile);
   magic[1] = getc(infile);
  }

 switch ( magic[1] )
  {
   case '1': /* pbm bitmap */
   case '4': /* binary */
      printf("ERROR: Bad PBM image file '%s'.\n", fname);
      return 0;
    break;
   case '2': /* pgm greyscale */
   case '5': /* binary */
   case '3': /* ppm pixmap */
   case '6': /* binary */
    while ( (ch = getc(infile)) && isspace(ch) );  /* Chew whitespace. */
    while ( ch == '#' ) 
     {
      while ( getc(infile) != '\n' );	/* Eat comment line. */
      ch = getc(infile);
     }
    ungetc( ch, infile );
    fscanf( infile, "%d", &ncols );	/* Number of columns */
    while ( (ch = getc(infile)) && isspace(ch) ); /* chew whitespace */
    while ( ch == '#' )
     {
      while ( getc(infile) != '\n' );	/* Eat comment line. */
      ch = getc(infile);
     }
    ungetc( ch, infile );
    fscanf( infile, "%d", &nrows ); 	/* Number of rows. */
    while ( (ch = getc(infile)) && isspace(ch) ); /* chew whitespace */
    while ( ch == '#' )
     {
      while ( getc(infile) != '\n' );	/* Eat comment line. */
      ch = getc(infile);
     }
    ungetc( ch, infile );
    fscanf( infile, "%d", &maxval );	/* Maximum value of a pixel. */
    while (getc(infile)!='\n');

    /* Get the data. */
    image = (struct Otk_image *)malloc( sizeof(struct Otk_image) );
    image->filename = (char *)strdup(ofname);
    image->image = (struct Otk_image_rec *)malloc( (ncols * nrows) * sizeof(struct Otk_image_rec) );
    if (image->image == 0) {printf("Otk Error: Could not allocate image memory.\n");  return 0; }
    image->native_cols = ncols;   image->native_rows = nrows;
    image->cols = ncols;   image->rows = nrows;

    if ( magic[1] == '6' ) 
     { /* binary */
      int k=0;
      for (row=0;  row != nrows;  row++)
       {
        for (col=0;  col != ncols;  col++)
         {
          fread( &rc, 1, 1, infile );
          fread( &gc, 1, 1, infile );
          fread( &bc, 1, 1, infile );
          image->image[k].r = rc;
          image->image[k].b = bc;
          image->image[k].g = gc;
	  k++;
         }
       }
     }
    else
    if ( magic[1] == '5' ) 
     { /* binary */
      int k=0;
      for (row=0;  row != nrows;  row++)
        for (col=0;  col != ncols;  col++)
        {
         fread( &rc, 1, 1, infile );
         image->image[k].r = rc;
         image->image[k].b = rc;
         image->image[k].g = rc;
	 k++;
        }
     }
    else
     { /* ascii */
      int k=0;
      for (row=0;  row != nrows;  row++)
       for (col=0;  col != ncols;  col++)
        {
         fscanf(infile,"%d",&r);    fscanf(infile,"%d",&g);     fscanf(infile,"%d",&b);
         image->image[k].r = r;
         image->image[k].b = b;
         image->image[k++].g = g;
        }
     }
    break;
  }
 if( ispipe ) pclose( infile );
 else    fclose( infile );
 // printf("nrows (columns) = %d,  ncols (rows) = %d\n", nrows, ncols );
 if (istmp) { remove("tmpimg.ppm");  free(fname); }
 return image;
}






void create_internal_image_data( char *imgfname, int quality, char *outfname, int count )
{ /* This routine writes static image declaration to file, for embedding into programs. */
  /* Quality: 3=maximum (18-bit color), 2=medium (15-bit color), 1=minimum (12-bit color), 0=maximum compression. */
 struct Otk_image *tmpimg;
 int ncols, nrows, col, row, j, k, mm=0;
 unsigned int last_r, last_g, last_b, rr, gg, bb;
 FILE *outfile;

 tmpimg = I2C_Read_Image_File( imgfname );
 if (tmpimg == 0) { printf("Error:  Cannot open image file '%s'\n", imgfname );  exit(1); }

 if (count == 0)
  outfile = fopen( outfname, "w" );
 else
  outfile = fopen( outfname, "a" );
 if (outfile == 0) {printf("Cannot open '%s'\n", outfname );  exit(1);}
 printf("Writing to '%s'\n", outfname );

 ncols = tmpimg->cols;
 nrows = tmpimg->rows;

 /* Decide if run-length encoding will pack this image-data smaller or larger than straight values. */
 k = 0;  j = 64;
 for (col=0; col < ncols; col++)
  {
   for (row=0; row < nrows; row++)
    {
     rr = tmpimg->image[mm].r >> 2;
     gg = tmpimg->image[mm].g >> 2;
     bb = tmpimg->image[mm].b >> 2;
     switch (quality)
      {
	case 2: rr = rr & 0xfffe;  gg = gg & 0xfffe;  bb = bb & 0xfffe;
		break;
	case 1: rr = rr & 0xfffc;  gg = gg & 0xfffc;  bb = bb & 0xfffc;
		break;
	case 0: rr = rr & 0xfff4;  gg = gg & 0xfff4;  bb = bb & 0xfff4;
		break;
      }
     if ((last_r - rr != 0) || (last_g - gg != 0) ||  (last_b - bb != 0) ||  (j >= 63))
      {
       last_r = rr;
       last_g = gg;
       last_b = bb;
       j = 0;
       k++;
      } 
     else j++;
     mm++;
    }
  }
 // printf("Straight coding would require %d bytes, while Run-Length coding will require %d bytes, or %2.2f%%.\n", 3 * ncols * nrows, 4 * k, 100.0 * (double)(4 * k)/(double)(3 * ncols * nrows) );

 if (3 * ncols * nrows <= 4 * k)
  { /*straight_coding*/
    fprintf(outfile,"\nint imgmtrx_N%d = %d;\n", count, 3 * ncols * nrows );
    /* Start with magic code that indicates the format being used (111). */
    fprintf(outfile,"\nunsigned char imgmtrxbytes%d[%d] = {\n\t\"111", count, 3 * ncols * nrows + 6 );
    k = 1;  mm = 0;
    for (col=0; col < ncols; col++)
     {
      for (row=0; row < nrows; row++)
       {
        if (k > 25) { fprintf(outfile,"\"\n\t\"");  k = 0; }
        fprintf(outfile,"%c%c%c", byte_encode( tmpimg->image[mm].r ), byte_encode( tmpimg->image[mm].g ), byte_encode( tmpimg->image[mm].b ) );
        k++;
        mm++;
      }
     }
    if (k > 25) fprintf(outfile,"\"\n\t\"");
    fprintf(outfile,"321\"\n");	/* Append magic end-marker as framing check. */
    fprintf(outfile,"\t};\n\n");
  } /*straight_coding*/
 else
  { /*run_length_code*/
    fprintf(outfile,"\nint imgmtrx_N%d = %d;\n", count, 4 * k );
    /* Start with magic code that indicates the format being used (112). */
    fprintf(outfile,"\nunsigned char imgmtrxbytes%d[%d] = {\n\t\"112", count, 4 * k + 6 );
    k = 1;  mm = 0;  j = 64;
    for (col=0; col < ncols; col++)
     {
      for (row=0; row < nrows; row++)
       {
	rr = tmpimg->image[mm].r >> 2;
	gg = tmpimg->image[mm].g >> 2;
	bb = tmpimg->image[mm].b >> 2;
	switch (quality)
	 {
	  case 2: rr = rr & 0xfffe;  gg = gg & 0xfffe;  bb = bb & 0xfffe;
		break;
	  case 1: rr = rr & 0xfffc;  gg = gg & 0xfffc;  bb = bb & 0xfffc;
		break;
	  case 0: rr = rr & 0xfff4;  gg = gg & 0xfff4;  bb = bb & 0xfff4;
		break;
	 }
	if ((last_r - rr != 0) || (last_g - gg != 0) ||  (last_b - bb != 0) ||  (j >= 63))
	 {
	  last_r = rr;
	  last_g = gg;
	  last_b = bb;
	  if (k > 25) { fprintf(outfile,"\"\n\t\"");  k = 0; }
	  if (j != 64) fprintf(outfile,"%c", byte_encode64( j ) );	/* Record the run-length of the previous value, if there was one. */
	  /* Now record the new rgb value. */				/* RL's range from 0 to 63, meaning RL's of 1 to 64. */ 
	  fprintf(outfile,"%c%c%c", byte_encode64( rr ), byte_encode64( gg ), byte_encode64( bb ) );
	  j = 0;
	  k++;
	 }
	else
	 j++;
        mm++;
      }
     }
    fprintf(outfile,"%c", byte_encode64( j ) );        /* Record the run-length of the last value. */
    if (k > 25) fprintf(outfile,"\"\n\t\"");
    fprintf(outfile,"321\"\n");	/* Append magic end-marker as framing check. */
    fprintf(outfile,"\t};\n\n");
  } /*run_length_code*/

 fprintf(outfile,"/* Use as:\n");
 fprintf(outfile,"	struct Otk_image *your_image%d;\n\n", count);
 fprintf(outfile,"	your_image%d = Otk_Make_Image_From_Matrix( \"YourImage%d\", %d, %d, decode_image( imgmtrx_N%d, %d, %d, imgmtrxbytes%d ) );\n", 
				count, count, nrows, ncols, count, nrows, ncols, count );
 fprintf(outfile,"	OtkMakeImagePanel_ImgPtr( OtkOuterWindow, your_image%d, 10, 10, 80, 80 );\n", count);
 fprintf(outfile,"*/\n\n");
 fclose( outfile );
}





main( int argc, char **argv )
{
 int argk=1, qual=3, count=0;
 char *outfname;

 printf("\nImage2Code - version 1.0\n");
 OtkInitWindow( 600, 360, 1, argv );
 outfname = strdup( "tmp_data.c" );

 /* First scan the command-line for any options. */
 while (argk < argc)
  {
   if (strcmp( argv[ argk ], "-o" ) == 0)
    {
	argk++;
	if (argk == argc) { printf("Missing file after '%s'.\n", argv[argk-1] );  exit(1); }
	outfname = strdup( argv[ argk ] );
    }
   else
   if (strcmp( argv[ argk ], "-qual" ) == 0)
    {
	argk++;
	if (argk == argc) { printf("Missing file after '%s'.\n", argv[argk-1] );  exit(1);}
	if (sscanf( argv[ argk ], "%d", &qual ) != 1)
	 { printf("Bad integer after '%s'.\n", argv[argk-1] );  exit(1);}
    }
   else
   if (strncmp( argv[ argk ], "-help", 2 ) == 0)
    {
	printf("Image2Code - Options:\n");
	printf("	-o xx	 - Send output to file xx instead of default 'tmp_data.c'.\n");
	printf("	-qual xx - Sets quality level to xx.  Allowable values range\n");
	printf("		   from 3 down to 0.  Default quality level is 3, which\n");
	printf("		   preserves good quality for all images with reasonable\n");
	printf("		   code size.  A value of 2 reduces code size, with\n");
	printf("		   usually little loss in image quality.  A value\n");
	printf("		   of 1 reduces code size further, and may be OK for\n");
	printf("		   line drawings.  Quality of zero produces the most\n");
	printf("		   compact code tables but image quality may suffer.\n\n");
	exit(0);
    }
   argk++;
  }

 /* Now get the file names. */
 argk = 1;
 while (argk < argc)
  {
   if (strcmp( argv[ argk ], "-o" ) == 0)
    {
	argk++;
    }
   else
   if (strcmp( argv[ argk ], "-qual" ) == 0)
    {
	argk++;
    }
   else
    {
     create_internal_image_data( argv[ argk ], qual, outfname, count++ );
				  /* Quality: 3=Good quality (18-bit color) OK for photo-images, */
				  /*	      2=Medium quality (15-bit color), 	*/
				  /*	      1=Lower quality (12-bit color) but good compression. */
    }
   argk++;
  }


 if (count == 0) printf("No image file names were on the command-line.\nNo output produced.\n\n" );
}
